#include <cstring>
#include <cstdio>
#include <cstdlib>

#include "configg.h"
#include "AuxFuncs.h"
#include "AuxRND.h"
#include "phong.h"

// This implements methods of a general interface

// Testing stochastic sampling and albedo/variance evaluation on phong model
//#define TEST_STOCHASTIC

// constructor
BRDF::BRDF():
{
  
}		      

// How to use probability for Metropolis sampling when it is stored to the array
// and when it should be used for the integratio as int=sum(importanceSampligValue/probValue)
float
funcProb(float newPri, float theta_view, float phi_view)
{
  const float epsMin = 1e-10f;
  float val = newPri + epsMin;
  assert(val > 0); // it cannot be zero - it makes a lot of problem
  return val;
}

// returns probability of sampling in a given viewer direction
float
BRDF::getProbOfSampling(int irow, int jcol,
			float theta_view, float phi_view,
			float theta_i, float phi_i,
			const float RGB[3]) const
{
  // general algorithm, if no other chance
  // value does exist as is provided by user
  float albedo = getAlbedo(irow, jcol, theta_view, phi_view);
  float prob = luminance(RGB)/albedo + 1e-25; // add small value to avoid zero probability
  return prob;
}

// Get albedo for viewing direction, computed numerically
float
BRDF::getAlbedoSampling(int irow, int jcol, // texture position
			// viewer direction as two angles in radians
			float theta_view, float phi_view) const
{
  // the three variables for stable incremental algorithm are reset
  individualSVA.reset();
  // generate samples just to compute albedo
  int cnt = cntThresholdMean - individualSVA.cnt;
  for (int i = 0; i < cnt; i++) {
    float theta_il, phi_il;
    // we compute 1/PI sampling over hemisphere here
    float rndxx = RandomValue();
    float rndyy = RandomValue();
    float sin_theta = sqrt(1.0000f - rndxx * rndxx);
    if (sin_theta >= 1.0f)
      sin_theta = 0.99999f;
    theta_il = asin(sin_theta);
    assert(theta_il >=0);
    assert(theta_il <= M_PI/2);
    phi_il = 2.0 * M_PI * rndyy;
    //return 1.0/(2.0f*M_PI); //probability of sampling over hemisphere for this direction
    float RGB[3];
    getBRDFValue(irow, jcol, // texture position
		 theta_view, phi_view, // viewer direction as two angles in radians
		 theta_il, phi_il, // illum. direction as two angles in radians
		 RGB); // output value
    // evaluate the function - this is the value of probability density function
    double newSampleProb = luminance(RGB);
    float cosTheta = cos(theta_il) + 1e-10f;
    assert(cosTheta > 0.f);
    float newSampleProb2 = newSampleProb/cosTheta;
    individualSVA.update(newSampleProb, newSampleProb2);
  } // for i

  // albedo is computed from series of luminance
  assert(individualSVA.cnt >= cntThresholdMean);
  individualSVA.evaluate();
  return individualSVA.albedo;
}

// Get albedo for viewing direction, computed numerically
void
BRDF::getAlbedoSamplingSimple(int irow, int jcol, // texture position
			      // viewer direction as two angles in radians
			      float theta_view, float phi_view,
			      SEntryVA &sva) const
{
  // the three variables for stable incremental algorithm are reset
  sva.reset();
  sva.irow = irow;
  sva.jcol = jcol;
  sva.theta = theta_view;
  sva.phi = phi_view;
  
  // generate samples just to compute albedo and variance/another statistics
  int cnt = cntThresholdMean;
  for (int i = 0; i < cnt; i++) {
    float theta_il, phi_il;
    // we compute 1/PI sampling over hemisphere here
    float rndxx = RandomValue();
    float rndyy = RandomValue();
    float sin_theta = sqrt(1.0000 - rndxx * rndxx);
    if (sin_theta >= 1.0f)
      sin_theta = 0.99999f;
    theta_il = asin(sin_theta);
    assert(theta_il >=0);
    assert(theta_il <= M_PI/2);
    phi_il = 2.0f * M_PI * rndyy;
    //return 1.0/(2.0f*M_PI); //probability of sampling over hemisphere for this direction
    float RGB[3];
    getBRDFValue(irow, jcol, // texture position
		 theta_view, phi_view, // viewer direction as two angles in radians
		 theta_il, phi_il, // illum. direction as two angles in radians
		 RGB); // output value
    // evaluate the function - this is the value of probability density function
    double lum = luminance(RGB);
    float cosTheta = cos(theta_il) + 1e-20f;
    assert(cosTheta > 0.f);
    float lum2 = lum/cosTheta;
    sva.update(lum, lum2);
  } // for i
  // albedo is computed from series of luminance
  assert(sva.cnt >= cntThresholdMean);
  sva.evaluate();
  return; // result
}

float
BRDF::getAlbedo(int irow, int jcol, // texture position
		// viewer direction as two angles in radians
		float theta_view, float phi_view) const
{
  // Here we return albedo either by interpolation or by sampling
  // from the BRDF. The computation is done in lazy way when needed
  return getAlbedoSampling(irow, jcol, theta_view, phi_view);
}

// Make stochastic sampling, where Z-axis is the normal of a surface
bool
BRDF::spawnStochasticRay(int irow, int jcol,
			 float rnd1, float rnd2, // input
			 float theta_view, float phi_view, // the direction of a viewer
			 float XYZ[3], // output direction
			 float RGB[3], // corresponding output color
			 float &prob) const // and probability for this direction
{
  // compute albedo and hence also the maximum value of the function
  float albedo = getAlbedoSampling(irow, jcol, theta_view, phi_view);
  // cout << "A = " << albedo << " maxlum = " << maxlum << endl;
  
  // generate samples to find out by rejection sampling acceptable sample
  bool acceptSample = false;
  int i = 0;
  for ( ;!acceptSample; ) {
    float theta_il, phi_il;
    // we compute 1/PI sampling over hemisphere here
    float rndxx = RandomValue();
    float rndyy = RandomValue();
    float sin_theta = sqrt(1.0000 - rndxx * rndxx);
    if (sin_theta >= 1.0f) sin_theta=0.99999;
    theta_il = asin(sin_theta);
    assert(theta_il >=0);
    assert(theta_il <= M_PI/2);
    phi_il       = 2.0 * M_PI * rndyy;
    //return 1.0/(2.0f*M_PI); //probability of sampling over hemisphere for this direction
    float RGB[3];
    getBRDFValue(irow, jcol, // texture position
		 theta_view, phi_view, // viewer direction as two angles in radians
		 theta_il, phi_il, // illum. direction as two angles in radians
		 RGB); // output value
    // evaluate the function - this is the value of probability density function
    double lum = luminance(RGB);
    float cosTheta=cos(theta_il)+1e-20;
    assert(cosTheta > 0.f);
    float lum2 = lum/cosTheta;
    individualSVA.update(lum, lum2);
    float rndaccept = RandomValue(); // between 0..1
    //cout << i++ << " " << maxlum << " " << lum << " " << maxlum * rndaccept << endl;
    if (individualSVA.maxlum2 * rndaccept < lum2) {
      // value generated is below the function
      //float albedo = 2.0f*M_PI*meanAlbedo;
      prob = lum; // value of probability density function
      ConvertThetaPhiToXYZ2(theta_il, phi_il, XYZ);
      acceptSample = true;
    }
  } // until a sample is found
  // we have found the sample
  individualSVA.evaluate();
  return false; // OK ... return
}

const char* 
BRDF::GetHelp() const
{
  return "Missing BRDF model description";
}

int 
BRDF::GetNumParameters() const
{
  return -1;
}

int
BRDF::GetParameter(int parameterIndex,) const
{
  return -1;
}

int
BRDF::GetParameterConstraints(int parameterIndex, float RGB[]) const
{
  return 0;
}

const char* 
BRDF::GetParameterDescription(int parameterIndex) const
{
  return "Missing parameter description";
}

BRDF::EParameterUIType
BRDF::GetParameterSyntax(int parameterIndex) const
{
  return EE_Unknown;
}

const char*
BRDF::GetParameterTitle(int parameterIndex) const
{
  return "???";
}

void
BRDF::SetParameter(int parameterIndex, float RGB[])
{
  
}


